-- This module is used by STC card for writing data to the LWB.
-- A 512*32 asynchronous FIFO is used to receive data from the STC.
-- Asynchronous FIFO is needed to allow STC use its own clock other
-- than the PCIBUS 2 clock for writing data to this module.
-- module reads data from the asfifo and recognizes the type of data.
-- A whole event is sent in one block starting from header to trailer
-- 12/9/03 count_h(z) added to count number of hit(z) event sent
-- hit is read back at offset 0x8
-- z is read back at offset 0xc
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;

entity pci2_if is
    port (
-- pci clock and reset signals
        clk: in STD_LOGIC;
        rstn: in STD_LOGIC;
-- pci function local side signals
        l_dato: in STD_LOGIC_VECTOR (31 downto 0);
        l_adro: in STD_LOGIC_VECTOR (31 downto 0);
        l_cmdo: in STD_LOGIC_VECTOR (3 downto 0);
        l_cbeni: out STD_LOGIC_VECTOR (3 downto 0);
        l_adi: out STD_LOGIC_VECTOR (31 downto 0);
-- pci function master local side signal
        stat_reg: in STD_LOGIC_VECTOR (5 downto 0);
        lm_tsr: in STD_LOGIC_VECTOR (9 downto 0);
        lm_dxfrn: in STD_LOGIC;
        lm_adr_ackn: in STD_LOGIC;
        lm_ackn: in STD_LOGIC;
        lm_rdyn: out STD_LOGIC;
        lm_req32n: out STD_LOGIC;
        lm_lastn: out STD_LOGIC;
-- pci function target local side signal
        lt_framen: in STD_LOGIC;
        lt_dxfrn: in STD_LOGIC;
        lt_ackn: in STD_LOGIC;
        lt_tsr: in STD_LOGIC_VECTOR (11 downto 0);
        lt_discn: out STD_LOGIC;
        lt_rdyn: out STD_LOGIC;
-- stc data
        stc_clk: in STD_LOGIC;
        stc_ctrl_i: in STD_LOGIC_VECTOR (3 downto 0);
        stc_ctrl_o: out STD_LOGIC_VECTOR (2 downto 0);
        led: out STD_LOGIC_VECTOR (3 downto 0);
        stc_data: in STD_LOGIC_VECTOR (31 downto 0)
    );
end pci2_if;

architecture pci2_arch of pci2_if is
COMPONENT lpm_counter
   GENERIC (LPM_WIDTH: POSITIVE := 8;
      LPM_MODULUS: NATURAL := 0;
      LPM_DIRECTION: STRING := "UNUSED";
      LPM_AVALUE: STRING := "UNUSED";
      LPM_SVALUE: STRING := "UNUSED";
      LPM_PVALUE: STRING := "UNUSED";
      LPM_TYPE: STRING := "LPM_COUNTER";
      LPM_HINT: STRING := "UNUSED");
   PORT (data: IN STD_LOGIC_VECTOR(LPM_WIDTH-1 DOWNTO 0) := (OTHERS => '0');
      clock: IN STD_LOGIC;
      cin: IN STD_LOGIC := '0';
      clk_en: IN STD_LOGIC := '1';
      cnt_en: IN STD_LOGIC := '1';
      updown: IN STD_LOGIC := '1';
      sload: IN STD_LOGIC := '0';
      sset: IN STD_LOGIC := '0';
      sclr: IN STD_LOGIC := '0';
      aload: IN STD_LOGIC := '0';
      aset: IN STD_LOGIC := '0';
      aclr: IN STD_LOGIC := '0';
      cout: OUT STD_LOGIC := '0';
      q: OUT STD_LOGIC_VECTOR(LPM_WIDTH-1 DOWNTO 0));
END COMPONENT;
COMPONENT lpm_fifo_dc
	GENERIC (LPM_WIDTH: POSITIVE := 32;
		LPM_WIDTHU: POSITIVE :=9;
		LPM_TYPE: STRING := "LPM_FIFO_DC";
		LPM_NUMWORDS: POSITIVE := 512;
		LPM_SHOWAHEAD: STRING := "OFF";
		LPM_HINT: STRING := "UNUSED");
	PORT (data: IN STD_LOGIC_VECTOR(LPM_WIDTH-1 DOWNTO 0);
		rdclock, wrclock, wrreq, rdreq : IN STD_LOGIC;
		aclr : IN STD_LOGIC := '0';
		rdfull, wrfull, wrempty, rdempty: OUT STD_LOGIC;
		rdusedw, wrusedw: OUT STD_LOGIC_VECTOR(LPM_WIDTHU-1 DOWNTO 0);
		q: OUT STD_LOGIC_VECTOR(LPM_WIDTH-1 DOWNTO 0)
		);
END COMPONENT;

signal stc_reset: std_logic;
signal lm_lastn_i: std_logic;
signal stc_data_r: std_logic_vector(31 downto 0);
signal stc_ctrl_i_r: std_logic_vector(3 downto 0);
signal toggle_h: std_logic;
signal toggle_z: std_logic;
signal toggle_h_sync: std_logic_vector(3 downto 0);
signal toggle_z_sync: std_logic_vector(3 downto 0);
signal we_fifo_h: std_logic;
signal we_fifo_z: std_logic;
signal re_fifo_h: std_logic;
signal re_fifo_z: std_logic;
signal wc_fifo_h: std_logic_vector(8 downto 0);
signal wc_fifo_z: std_logic_vector(8 downto 0);
signal en_h: std_logic;
signal en_z: std_logic;
signal empty_fifo_h: std_logic;
signal empty_fifo_z: std_logic;
signal h_data_valid: std_logic;
signal z_data_valid: std_logic;
signal h_got_event: std_logic;
signal z_got_event: std_logic;
signal h_empty: std_logic;
signal z_empty: std_logic;
signal ce_fifo_ec_h: std_logic;
signal ce_fifo_ec_z: std_logic;
signal fifo_ec_h: std_logic_vector(4 downto 0);
signal fifo_ec_z: std_logic_vector(4 downto 0);
signal h_data: std_logic_vector(31 downto 0);
signal z_data: std_logic_vector(31 downto 0);
signal sel_z: std_logic;
signal prio_z: std_logic;
signal idle: std_logic;
signal ltrdyn: std_logic;
--signal sel_fifo: std_logic;
signal fifo_data: std_logic_vector(31 downto 0);
signal ltb_addr_h: std_logic_vector(31 downto 7);
signal ltb_addr_z: std_logic_vector(31 downto 7);
signal ltb_addr: std_logic_vector(31 downto 0);
signal ec_ec: std_logic;
signal ec: std_logic_vector(4 downto 0);
-- debugging signal
signal en_count_h : std_logic;
signal en_count_z : std_logic;
signal count_h: std_logic_vector(31 downto 0);
signal count_z: std_logic_vector(31 downto 0);
signal counts: std_logic_vector(31 downto 0);
signal header_h: std_logic_vector(31 downto 0);
signal trailer_h: std_logic_vector(31 downto 0);
signal got_cnt: std_logic_vector(31 downto 0);
signal sent_cnt: std_logic_vector(31 downto 0);
signal singlet_o: std_logic_vector(31 downto 0);
signal en_header_h : std_logic;
signal en_trailer_h : std_logic;
signal lm_adr_ackn_dl : std_logic;
signal ec_sent_cnt : std_logic;
signal en_singlet_o : std_logic;
signal was_trailer_i : std_logic;
signal first_data : std_logic;
signal latch_stat : std_logic;
signal stat_reg_l: std_logic_vector(5 downto 0);
begin
sel_z <= '0';
led <= "0000";
stc_ctrl_o(2) <= '0';
stc_ctrl_o(1) <= '0';
stc_reset <= not stc_ctrl_i_r(3);
lm_lastn <= lm_lastn_i;
-- stc_ctrl_o(0) = '1' flags hit fifo almost full
-- stc_ctrl_o(1) = '1' flags z fifo almost full
process(stc_clk)
begin
  if(stc_clk'event and stc_clk = '1')then
-- buffer inputs to improve timing
    stc_data_r <= stc_data;
    stc_ctrl_i_r <= stc_ctrl_i;
-- fifo almost full signals
    if(wc_fifo_h(8 downto 5) = "1111")then
      stc_ctrl_o(0) <= '1';
    else
      stc_ctrl_o(0) <= '0';
    end if;
  end if;
end process;
i_fifo_h: lpm_fifo_dc
       port map(data => stc_data_r,
                rdclock => clk,
                wrclock => stc_clk,
                wrreq => we_fifo_h,
                rdreq => re_fifo_h,
                aclr => stc_reset,
                rdempty => empty_fifo_h,
                wrusedw => wc_fifo_h,
				q => h_data);
we_fifo_h <= not stc_ctrl_i_r(1) and not stc_ctrl_i_r(0) and en_h;
-- fifo empty is checked by the LPM
re_fifo_h <= not h_data_valid or not lm_dxfrn; 
process(stc_clk,stc_reset)
begin
  if(stc_reset = '1')then
    toggle_h <= '0';
  elsif(stc_clk'event and stc_clk = '1')then
	if(stc_data_r(31 downto 28) = "1111" and we_fifo_h = '1')then
	  toggle_h <= not toggle_h;
	end if;
  end if;
end process;
process(clk,stc_reset)
begin
  if(stc_reset = '1')then
    toggle_h_sync <= "0000";
    h_got_event <= '0';
    h_data_valid <= '0';
  elsif(clk'event and clk = '1')then
    toggle_h_sync <= toggle_h_sync(2 downto 0) & toggle_h;
	h_got_event <= toggle_h_sync(3) xor toggle_h_sync(2);
	h_data_valid <= not empty_fifo_h or (h_data_valid and lm_dxfrn); 
  end if;
end process;
i_ec: lpm_counter
   generic map(LPM_WIDTH => 5)
   port map(
           clock => clk,
           clk_en => ec_ec,
           updown => h_got_event,
           aclr => stc_reset,
           q => ec(4 downto 0));
ec_ec <= h_got_event xor (lm_adr_ackn_dl and not lm_adr_ackn);
i_got_event: lpm_counter
   generic map(LPM_WIDTH => 8)
   port map(
           clock => clk,
           clk_en => h_got_event,
           aclr => stc_reset,
           q => got_cnt(7 downto 0));
i_sent_event: lpm_counter
   generic map(LPM_WIDTH => 8)
   port map(
           clock => clk,
           clk_en => ec_sent_cnt,
           aclr => stc_reset,
           q => sent_cnt(7 downto 0));
process(clk)
begin
  if(clk'event and clk = '1')then
    lm_adr_ackn_dl <= lm_adr_ackn;
  end if;
end process;
ec_sent_cnt <= lm_adr_ackn_dl and not lm_adr_ackn;
-- if not in any of the active phases, master is in idle
idle <= not (lm_tsr(0) or lm_tsr(1) or lm_tsr(2) or lm_tsr(3));
-- request bus
lm_req32n <= '0' when idle = '1' and ec /= "00000" else '1';
--
lm_lastn_i <= '0' when lm_dxfrn = '0' and fifo_data(31 downto 28) = "1111" else '1'; 
-- master always writes Double Word
l_cbeni <= "0111" when lm_adr_ackn = '0' else "1111";
fifo_data <= h_data;
l_adi <= ltb_addr when lm_adr_ackn = '0' or (lt_tsr(0) = '1' and l_adro(4 downto 3) = "00") else counts when lt_tsr(0) = '1' else
         fifo_data;
--counts <= count_z when l_adro(2) = '1' else count_h;
process(l_adro,got_cnt,count_h,header_h,trailer_h,sent_cnt,singlet_o)
begin
  case l_adro(4 downto 2) is
    when "010" => counts <= count_h;
    when "011" => counts <= singlet_o;
    when "100" | "101" => counts <= header_h;
--    when "101" => counts <= trailer_h;
    when "110" => counts <= got_cnt;
    when others => counts <= sent_cnt;
  end case;
end process;
process(stc_clk,stc_reset)
begin
  if(stc_reset = '1')then
    was_trailer_i <= '1';
  elsif(stc_clk'event and stc_clk = '1')then
    was_trailer_i <= (was_trailer_i and (stc_ctrl_i_r(1) or stc_ctrl_i_r(0))) or en_trailer_h;
  end if;
end process;
--header_h(31 downto 24) <= "00000000";
--trailer_h(31 downto 24) <= "00000000";
count_h(31 downto 24) <= "00000000";
got_cnt(31 downto 8) <= "000000000000000000000000";
sent_cnt(31 downto 8) <= "000000000000000000000000";
singlet_o(31 downto 8) <= "000000" & lm_tsr & empty_fifo_h & h_data_valid & stat_reg_l;
process(clk,stc_reset)
begin
  if(stc_reset = '1')then
    stat_reg_l <= "000000";
  elsif(clk'event and clk = '1')then
    stat_reg_l <= stat_reg_l or stat_reg;
  end if;
end process;
-- now it is total word count
i_header_h: lpm_counter
   generic map(LPM_WIDTH => 32)
   port map(
--           clock => stc_clk,
           clock => clk,
           clk_en => en_header_h,
           aclr => stc_reset,
           q => header_h);
--en_header_h <= '1' when was_trailer_i = '1' and stc_ctrl_i_r(1 downto 0) = "00" and stc_data_r(31 downto 28) = "1100" else '0';
en_header_h <= not lm_dxfrn;
i_trailer_h: lpm_counter
   generic map(LPM_WIDTH => 32)
   port map(
           clock => stc_clk,
           clk_en => en_trailer_h,
           aclr => stc_reset,
           q => trailer_h);
--en_trailer_h <= '1' when stc_ctrl_i_r(1 downto 0) = "00" and stc_data_r(31 downto 28) = "1111" else '0';
en_trailer_h <= we_fifo_h;
i_singlet_o: lpm_counter
   generic map(LPM_WIDTH => 8)
   port map(
           clock => clk,
           clk_en => en_singlet_o,
           aclr => stc_reset,
           q => singlet_o(7 downto 0));
en_singlet_o <= '1' when first_data = '1' and lm_dxfrn = '0' and h_data(31 downto 28) = "1111" else '0';
i_count_h: lpm_counter
   generic map(LPM_WIDTH => 24)
   port map(
           clock => clk,
           clk_en => en_count_h,
           aclr => stc_reset,
           q => count_h(23 downto 0));
en_count_h <= not lm_lastn_i;
process(clk)
begin
  if(clk'event and clk = '1')then
    first_data <= not lm_adr_ackn or (first_data and lm_dxfrn);
  end if;
end process;
process(clk)
begin
  if(clk'event and clk = '1')then
    if(ltb_addr_h(31 downto 7) = "0000000000000000000000000")then
      en_h <= '0';
    else
      en_h <= '1';
    end if;
    if(ltb_addr_z(31 downto 7) = "0000000000000000000000000")then
      en_z <= '0';
    else
      en_z <= '1';
    end if;
--    sel_fifo <= not lm_adr_ackn or (sel_fifo and (lm_tsr(2) or lm_tsr(3)));
  end if;
end process;
-- output data mux
ltb_addr(31 downto 7) <= ltb_addr_z when sel_z = '1' or (lt_tsr(0) = '1' and l_adro(2) = '1') else ltb_addr_h;
--ltb_addr(31 downto 7) <= ltb_addr_h;
ltb_addr(6 downto 0) <= "1000000" when lm_adr_ackn = '0' else "0000000";
-- master always ready when it has the bus
--lm_rdyn <= not ((lm_tsr(1) or lm_tsr(2) or lm_tsr(3)) and lm_lastn_i);
lm_rdyn <= '1' when lm_tsr(3 downto 2) = "00" else '0';
lt_rdyn <= ltrdyn;
lt_discn <= (ltrdyn or not l_cmdo(0)) and lt_dxfrn;
process(clk,rstn)
begin
  if (rstn = '0')then
    ltb_addr_h <= (others => '0');
    ltb_addr_z <= (others => '0');
    ltrdyn <= '1';
  elsif(clk'event and clk = '1')then
-- target function has only one 32 bit register mapped to memory space.
-- It saves the LTB's base address and enables the master
    if(lt_dxfrn = '0' and l_cmdo(0) = '1' and l_adro(2) = '0')then
      ltb_addr_h <= l_dato(31 downto 7);
    end if;
    if(lt_dxfrn = '0' and l_cmdo(0) = '1' and l_adro(2) = '1')then
      ltb_addr_z <= l_dato(31 downto 7);
    end if;
-- target is always ready and accepts only single cycle access
    ltrdyn <= lt_framen;
-- fifo output validity flags
  end if;
end process;
end pci2_arch;

